#include <cmath>
#include <msp/core/algorithm.h>
#include <msp/core/maputils.h>
#include <msp/core/raii.h>
#include "animate.h"
#include "fadeoverlay.h"
#include "sequencer.h"
#include "stage.h"

using namespace std;

namespace Msp {
namespace DemoScene {

Sequencer::Sequencer(float bpm):
	started(false),
	beat(0),
	next_event(0)
{
	set_beats_per_minute(bpm);

	register_action_type<Animate>("animate");
	register_action_type<FadeOverlay::Fade>("fade");
	register_action_type<Stage::UseInView>("use_stage");
	register_action_type<Stage::SetCamera>("set_camera");
}

void Sequencer::set_beats_per_minute(float bpm)
{
	secs_per_beat = Time::min/bpm;
}

void Sequencer::add_static_action(Action &act)
{
	static_actions.push_back(&act);
}

void Sequencer::add_action(Action &act, float sb, float eb)
{
	if(sb<0 || sb>eb)
		throw invalid_argument("Sequencer::add_action");

	Segment seg;
	seg.action = &act;
	seg.start_beat = sb;
	seg.end_beat = eb;
	auto i = find_if(segments, [=](const Segment &s){ return s.start_beat>sb; });
	segments.insert(i, seg);
}

void Sequencer::start()
{
	started = true;
	end = begin = segments.begin();
	for(auto a: static_actions)
		a->start(0, numeric_limits<float>::max());
	float start_beat = beat;
	beat = -1;
	advance_to(start_beat);
}

void Sequencer::seek(float b)
{
	if(b<beat)
		throw logic_error("Sequencer::seek");

	if(started)
		advance_to(b);
	else
		beat = b;
}

void Sequencer::tick(const Time::TimeDelta &dt)
{
	if(!started)
		start();
	if(begin==segments.end())
		return;

	advance_to(beat+dt/secs_per_beat);
}

void Sequencer::advance_to(float target_beat)
{
	while(target_beat>next_event)
		advance_to(next_event);

	float prev_beat = beat;
	beat = target_beat;

	while(end!=segments.end() && end->start_beat<=beat)
		++end;

	int ibeat = static_cast<int>(floor(beat));
	bool do_beat = (ibeat!=static_cast<int>(floor(prev_beat)));

	for(auto a: static_actions)
	{
		if(do_beat)
			a->beat(ibeat);
		a->tick(beat, beat-prev_beat);
	}

	for(auto i=begin; i!=end; ++i)
	{
		if(i->start_beat>prev_beat)
			i->action->start(i->start_beat, i->end_beat-i->start_beat);
		if(do_beat)
			i->action->beat(ibeat);
		if(i->end_beat>=prev_beat)
		{
			float tick_beat = min(beat, i->end_beat);
			i->action->tick(tick_beat, tick_beat-max(prev_beat, i->start_beat));
			if(i->end_beat<=beat)
				i->action->end(i->end_beat);
		}
	}

	while(begin!=end && begin->end_beat<=beat)
		++begin;

	if(target_beat>=next_event)
		update_next_event();

	if(begin==segments.end())
		signal_finished.emit();
}

void Sequencer::update_next_event()
{
	next_event = numeric_limits<float>::max();

	if(end!=segments.end())
		next_event = min(next_event, end->start_beat);

	for(auto i=begin; i!=end; ++i)
		if(i->end_beat>beat)
			next_event = min(next_event, i->end_beat);
}


Sequencer::Loader::Loader(Sequencer &s, Demo &d):
	DataFile::ObjectLoader<Sequencer>(s),
	demo(d),
	base_beat(0.0f)
{
	add("base", &Loader::base);
	add("define_action", &Loader::define_action);
	add("instant", &Loader::instant);
	add("repeat", &Loader::repeat);
	add("segment", &Loader::segment);
}

void Sequencer::Loader::base(float b)
{
	SetForScope<float> set_base(base_beat, base_beat+b);
	load_sub_with(*this);
}

void Sequencer::Loader::define_action(const string &n)
{
	ActionDefLoader ldr(obj, demo);
	load_sub_with(ldr);
	obj.named_actions[n] = ldr.get_action();
}

void Sequencer::Loader::instant(float beat)
{
	segment(beat, beat);
}

void Sequencer::Loader::repeat(float b, float d, unsigned n)
{
	for(unsigned i=0; i<n; ++i)
	{
		SetForScope<float> set_base(base_beat, base_beat+b+i*d);
		load_sub_with(*this);
	}
}

void Sequencer::Loader::segment(float start, float end)
{
	SegmentLoader ldr(obj, base_beat+start, base_beat+end, demo);
	load_sub_with(ldr);
}


Sequencer::ActionDefLoader::ActionDefLoader(Sequencer &s, Demo &d):
	DataFile::ObjectLoader<Sequencer>(s),
	demo(d)
{
	obj.action_registry.add_all(*this);
}


Sequencer::SegmentLoader::SegmentLoader(Sequencer &s, float b, float e, Demo &d):
	ActionDefLoader(s, d),
	start_beat(b),
	end_beat(e)
{
	add("apply", &SegmentLoader::apply);
}

void Sequencer::SegmentLoader::action_loaded()
{
	obj.add_action(*action, start_beat, end_beat);
	obj.anonymous_actions.push_back(action.release());
}

void Sequencer::SegmentLoader::apply(const string &n)
{
	obj.add_action(*get_item(obj.named_actions, n), start_beat, end_beat);
}

} // namespace DemoScene
} // namespace Msp
